#!/usr/bin/perl

# This code is (C) 2002-2006 by Pawel T. Jochym
# and it is licensed to use and redistribute under 
# the terms of the GNU General Public License v2
# See http://www.gnu.org/licenses/licenses.html 
# for the details of the terms and conditions.
 
$id='$Id: bp,v 1.7 2002/06/20 12:38:04 jochym Exp $';
$version='$Revision: 1.7 $';

use POSIX ":sys_wait_h"; 


$verb=1;
$c2out=1;
$debug=0;
$RADS=3.141592/180;

# Calculation parameters
$cutoff='MEDIUM';
$rmsforce=0.001;
$rmsstress=0.05;
$runmode="NQS";
$VASPNAME="vasp";

# INCAR params
$NELM=60;
$EDIFF=1.E-8;
$EDIFFG=-1.E-5;
$EDIFFG=1.E-7;
$IBRION=1;
$NSW=30;
$NSW_DFRM=30;
$ADDGRID=".FALSE.";
$CUSTOM_INCAR="";

print STDERR "BvsP $version\n";
print STDERR "$id\n" if $verb;

$inpipe="datain.$$";
$castep="/disk2/usr/people/bajochym/bin/cerius38";
$homedir="$ENV{HOME}";
$statfile="$homedir/.MSIcastepstat";
#$statfile=".rj.remote";
$modelBase=$ARGV[0];
$model="./$modelBase.msi";
%lattices = (
    "triclinic" => 18 ,
    "monoclinic" => 12 ,
    "orthorombic" => 9 ,
    "orthorhombic" => 9 ,
    "tetragonal" => 6 ,
    "rhombohedric" => 6 ,
    "rhomboedric" => 6 ,
    "hexagonal" => 5 ,
    "cubic" => 3 
    );
%deformations = ( 
    cubic => [ UNIAXIAL_A, SHEER_A ],
    orthorombic => [ UNIAXIAL_A, UNIAXIAL_B, UNIAXIAL_C,
			SHEER_A, SHEER_B, SHEER_C ],
    orthorhombic => [ UNIAXIAL_A, UNIAXIAL_B, UNIAXIAL_C,
			SHEER_A, SHEER_B, SHEER_C ],
    rhombohedric => [ UNIAXIAL_A, UNIAXIAL_B, UNIAXIAL_C,
			SHEER_A, SHEER_B, SHEER_C ],
    rhomboedric => [ UNIAXIAL_A, UNIAXIAL_B, UNIAXIAL_C,
			SHEER_A, SHEER_B, SHEER_C ],
    hexagonal => [ UNIAXIAL_A, UNIAXIAL_B, UNIAXIAL_C,
			SHEER_A, SHEER_B, SHEER_C ],
    tetragonal => [ UNIAXIAL_A, UNIAXIAL_C, SHEER_B, SHEER_C ],
    monoclinic => [ UNIAXIAL_A, UNIAXIAL_B, UNIAXIAL_C,
			SHEER_A, SHEER_B, SHEER_C ],
    );
%anglnam = (
	    A => ALPHA,
	    B => BETA,
	    C => GAMMA
	    );

@pressures = (-2, -1, 0, 1, 2, 4, 10, 20, 40, 80);
@angles = ( 0, 1, 2 );
@shrinks = ( -1, 0, 1 );

$intdofsetup="EDIT/MOVEABILITY MOVEABLE";

# now override with local settings
require "./PARAMS";
    
sub dprn
{
    print STDERR @_ if $debug;
}

sub prn
{
    print STDERR @_ ;
}

sub vprn
{
    print STDERR @_ if $verb;
}

sub cleanupSIG
{
    vprn "Cleaning up ...\n" if $verb;
    vprn "Killing VASP process $VPID\n";
    kill 9, $VPID;
    exit 0;
}

sub cleanup
{
    print STDERR "Cleaning up ...\n" if $verb;
}

# Mixed product of cell vectors (cell volume)
sub vol
{
    my @c=@_;
    my $v=0;
    
    $v+=$c[6]*($c[1]*$c[5]-$c[2]*$c[4]);
    $v+=$c[7]*($c[2]*$c[3]-$c[0]*$c[5]);
    $v+=$c[8]*($c[0]*$c[4]-$c[1]*$c[3]);
    $v;
}

sub analysePvsV
{
    my ($fname)=@_;
    my ($p, $v);
    
    open(CST,$fname) or die "Cannot open $fname: $!\n";

    vprn "Reading $fname ... \n";
    
    while (<CST>) {
	if (/^ +external pressure = +([0-9.-]+) kB +Pullay stress = +([0-9.-]+) kB/) {
	    $p=$1+$2;
	}
	if (/^ +volume of cell : +([0-9.]+)/) {
	    $v=$1;
	}
	if (/ +TOTEN  = +([-0-9.]+) eV/) {
	    $e=$1;
	}
    }
    close(CST);
    $PvsV{$v}="$p:$e";
}

sub analyseModelLoad
{
    while(<CERIUS>) {
	if (/Lattice type: +([-A-Z\.]+)/) {
	    $lattice_type=$1;
	}
	if (/Lattice centering: +([-A-Z\.]+)/) {
	    $lattice_centering=$1;
	}
	last if (/^ +Total number of general positions/);
    }
}

sub readModelCell_from_OUTCAR
{
    my ($name)=@_;
    
    open(MDL,"$name/OUTCAR") or 
	die "Cannot open model file $name/OUTCAR : $!\n";
    
    vprn "Reading model from OUTCAR: $name ...\n" ;
    
    while (<MDL>) {
	if (/LATTYP: Found a.* \(*([\w]+)\)* cell./) {
	    $lattice_type=$1;
	    vprn "Lattice type: $lattice_type\n";
	}
	if (/A([123]) = \( +([0-9.E\-+]+), +([0-9.E\-+]+), +([0-9.E\-+]+)\)/) {
	    dprn "$2 $3 $4\n" ;
	    if ($1 eq "1") {@a=($2,$3,$4)};
	    if ($1 eq "2") {@b=($2,$3,$4)};
	    if ($1 eq "3") {@c=($2,$3,$4); last };
	}
    }
    close(MDL);
    
    @cellv=(@a, @b, @c);
    @cell_dfrm=@cellv;
    ($A,$B,$C)=@cell=(sqrt($a[0]**2+$a[1]**2+$a[2]**2),
		      sqrt($b[0]**2+$b[1]**2+$b[2]**2), 
		      sqrt($c[0]**2+$c[1]**2+$c[2]**2));
	
    vprn "Cell: [$A, $B ,$C] V=",vol(@cellv),"\n";
}

sub readModelCell_from_CONTCAR
{
    my ($name)=@_;
    my $s=0;
    my $scale;
    
    open(MDL,"$name/CONTCAR") or 
	die "Cannot open model file $name/CONTCAR : $!\n";
    
    vprn "Reading model from CONTCAR: $name ...\n" ;
    
    while (<MDL>) {
	if ($s==0) {
	    $s++;
	    vprn "System: $_\n";
	}
	elsif ($s==1 && /^ *([-0-9.eE]+)/) {
	    $scale=$1;
	    $s++;
	    vprn "Lattice scale: $scale\n";
	}
	if ($s>=2 && /^ *([0-9.E\-+]+) +([0-9.E\-+]+) +([0-9.E\-+]+)/) {
	    dprn "$1 $2 $3\n" ;
	    $s++;
	    if ($s == 3) {@a=($1*$scale, $2*$scale,$3*$scale)};
	    if ($s == 4) {@b=($1*$scale, $2*$scale,$3*$scale)};
	    if ($s == 5) {@c=($1*$scale, $2*$scale,$3*$scale); last };
	}
    }
    close(MDL);
    
    @cellv=(@a, @b, @c);
    @cell_dfrm=@cellv;
    ($A,$B,$C)=@cell=(sqrt($a[0]**2+$a[1]**2+$a[2]**2),
		      sqrt($b[0]**2+$b[1]**2+$b[2]**2), 
		      sqrt($c[0]**2+$c[1]**2+$c[2]**2));
	
    vprn "Cell: [$A, $B ,$C] V=",vol(@cellv),"\n";
}

sub readModelCell
{
    my ($name)=@_;
    readModelCell_from_OUTCAR($name);
    readModelCell_from_CONTCAR($name);
}

sub runVASP
{
    my ($d,$w)=@_;
    my ($nel,$nion,$pnel,$pnion,$time);
    if (-e "$d/finished") {
	prn "VASP calculation ready in this directory.\n";
	prn "\n!!!!!!!!!\n\nTHIS SHOULD NOT BE SO!\n CONTACT AUTHOR, SOMETHING IS WRONG\n";
	prn "The runVASP procedure was probably called directly instead of RunCalc\n";
    } else {
	dprn "Running VASP in $d ... \n";
	unlink("$d/finished");
	# Simulation ...
	if ($w) {
	    system("cd $d ; $VASPNAME & ");
	    sleep 5;
	    $nel=$nion=$pnel=$pnion=0;
	    vprn "Waiting for VASP(*/10s):";
	    sleep 10;
	    $time=6;
	    until(-e "$d/calculating") {
		vprn "*";
		sleep 10;
		die "\nTimeout: VASP not started, exiting \n" if ($time-- < 0);
	    }
	    ($VPID,@dum)=split ' ',`cat "$d/calculating"`;
	    vprn " started with PID=$VPID\n";
    	    until( -e "$d/finished") {
		if ($verb > 32) {
		    ($nel,$dum)=split ' ',`wc -l "$d/OSZICAR"`;
		    ($nion,$dum)=split ' ',`grep F= "$d/OSZICAR" | wc -l`;
		    if ($nion>$pnion) { 
	    		vprn "o" x $nion, " " x ($nel-$pnel),"\r"; 
			$pnion=$nion;
			$pnel=$nel; 
		    }
		    vprn "o" x $nion, "." x ($nel-$pnel),"\r";
		    #vprn -e $d."/finished" ;
		}
		sleep 5;
	    }
	    vprn " $VPID job in $d finished\n ";
	} else {
	    system("cd $d ; $QVASPNAME & ");
	};
    }
}

# No check of existence of directory structure
sub createCalcDir
{
    my ($d,$p,$suf,$type) = @_;
    $destdir="$d"."_$p";
    mkdir "$destdir", 0777 ;
    dprn "$destdir\n" ;
    $destdir="$destdir/$destdir"."_$suf";
    dprn "$destdir\n" ;
    mkdir "$destdir", 0777;
    system("touch $destdir/calculating");
    symlink "../../$d/KPOINTS", "$destdir/KPOINTS";
    symlink "../../$d/POTCAR", "$destdir/POTCAR";
    symlink "../../$d/vasprun.conf", "$destdir/vasprun.conf";
    $nsw=$NSW;
    if ($type eq "Cell") {
	$ISIF=6;
	symlink "../../$d/POSCAR", "$destdir/POSCAR";
	symlink "../$destdir", "$destdir/../Cell";
    } elsif ($type eq "IDOF") {
	$ISIF=2;
	symlink "../Cell/CONTCAR", "$destdir/POSCAR";
	symlink "../$destdir", "$destdir/../IDOF";
    } elsif ($type eq "DFRM") {
	$ISIF=2;
	$nsw=$NSW_DFRM;
	symlink "../$destdir", "$destdir/../$suf";
	$CONTCAR="$destdir/../IDOF/CONTCAR";
	open CONTCAR or die "Cannot open $POSCAR . Exiting\n";
	$POSCAR=">$destdir/POSCAR";
	open POSCAR or die "Cannot open $POSCAR . Exiting\n";
	$n=0;
	while (<CONTCAR>) {
	    print POSCAR $_ if ($n<2 || $n>4);
	    if ($n>1 && $n<5) {
		printf POSCAR "    %25.20f   %25.20f   %25.20f\n",
		    $cell_dfrm[($n-2)*3+0],
		    $cell_dfrm[($n-2)*3+1],
		    $cell_dfrm[($n-2)*3+2];
	    }
	    $n++;
	}
	close POSCAR;
	close CONTCAR;
    } elsif ($type eq "SP") {
	$ISIF=7;
	symlink "../IDOF/CONTCAR", "$destdir/POSCAR";
	$nsw=0;
    }

    $INCAR=">$destdir/INCAR";
    open INCAR or die "Cannot create $INCAR\n";
    print INCAR <<EOF;
SYSTEM = $d, $p, $suf opt.
PREC = High
NELM = $NELM
EDIFF = $EDIFF 
EDIFFG = $EDIFFG
IBRION = $IBRION
ISIF = $ISIF
NSW = $nsw
PSTRESS = $p
LWAVE = .FALSE.
LCHARG = .FALSE.
ADDGRID= $ADDGRID
$CUSTOM_INCAR
EOF
    close INCAR;
    $destdir;
}

sub RunCalc
{
    my ($d,$p,$suf,$type) = @_;
    my $cd;
    $destdir="$d"."_$p";
    if ( -e "$destdir/$destdir"."_$suf/finished" ) {
	prn "$type : VASP calculation ready, skipping.\n";
	return "$destdir/$destdir"."_$suf";
    }
    $cd=createCalcDir($d,$p,$suf,$type);
    runVASP($cd,(($type eq "Cell") or ($type eq "IDOF")));
    $cd;  
}

# Run sequence of VASP runs for pressure $p 
sub runVASPforP
{
    my ($p)=@_;


    prn "Deformations for P=$p\n" ;
    
    
    readModelCell($modelBase);
    
# Set up calculations

# Optimize unit cell for geometry - this should be done manually

# Load model optimized for P=0 
    
# The following three steps are done conditionally
# if there are no pre-calculated directories
# the conditioning is embeded in RunCalc
# For each external pressure optimize unit cell
# Internal DOF are fixed
    prn "Cell optimization\n";
    $dc=RunCalc($modelBase,$p,"Cell","Cell");

# For each cell geometry relax internal DOF
# Cell vectors are fixed
    prn "IDOF optimization\n";
    $di=RunCalc($modelBase,$p,"IDOF","IDOF");
    
    analysePvsV "$di/OUTCAR";
    
# Calculate stress tensor for optimized geometry
#    $ds=RunCalc($modelBase,$p,"SP","SP");

# For each pressure cell geometry relax internal DOF
# after deforming it as needed by cell symetry
# Cell vectors are fixed
    if (-e "$di/finished" and -e "$dc/finished") {
	readModelCell($di);
	$dl=$deformations{$lattice_type};
#	prn join(", ",@$dl);
	$dln=$#{$dl};
	foreach $d (0 .. $dln) {
	    $dfm=${$dl}[$d];
	    prn "Calculating stress tensor for ${$dl}[$d] at $p kB\n" ;
	    if ( $dfm =~ /SHEER_([A-C])/ ) {
		foreach $ds ( @angles ) {
		    prn "Deformation of $anglnam{$1} by $ds deg.\n" ;
		    vprn "WARNING only implemented for orthogonal cells.\n" ;
		    @cell_dfrm=@cellv;
		    $eps=sin($ds*$RADS)/cos($ds*$RADS);
		    if ($1 eq "A") {
			$cell_dfrm[1]+=$eps*$cellv[2];
			$cell_dfrm[4]+=$eps*$cellv[5];
			$cell_dfrm[7]+=$eps*$cellv[8];
		    } elsif ($1 eq "B") {
			$cell_dfrm[0]+=$eps*$cellv[2];
			$cell_dfrm[3]+=$eps*$cellv[5];
			$cell_dfrm[6]+=$eps*$cellv[8];
		    } else { # "C"
			$cell_dfrm[0]+=$eps*$cellv[1];
			$cell_dfrm[3]+=$eps*$cellv[4];
			$cell_dfrm[6]+=$eps*$cellv[7];
    		    }
		    dprn "Cell Volume: ",vol(@cellv)," -> ",vol(@cell_dfrm),"\n";
		    $d=RunCalc($modelBase,$p,"S$1_${ds}","DFRM");
		}
	    }
	    if ( $dfm =~ /UNIAXIAL_([A-C])/ ) {
		foreach $ds ( @shrinks ) {
#		    next if (-e "${modelBase}");
		    $CV=$cell[(ord($1)-ord('A'))];
		    $CV*=(1+$ds/100);
		    $dfrm_cell[(ord($1)-ord('A'))]=$CV;
		    prn "Deformation of $1 by $ds%\n" ;
		    @cell_dfrm=@cellv;
		    $nr=(ord($1)-ord('A'));
		    $cell_dfrm[$nr]*=(1+$ds/100);
		    $cell_dfrm[$nr+3]*=(1+$ds/100);
		    $cell_dfrm[$nr+6]*=(1+$ds/100);
		    $d=RunCalc($modelBase,$p,"U$1_${ds}","DFRM");
		}
	    }
	}
    } else {
    	print STDERR "WARNING !!!! No optimized geometry for $p kB!\n";
	print STDERR "No calculation of stress tensor for $p kB\n";
    }
    

}


# MAIN part

$SIG{INT}='cleanupSIG';
$SIG{QUIT}='cleanupSIG';

# Do runs for each pressure
my(@kids);

foreach $P (@pressures) {
    FORK: {
	if (my $pid = fork) {
	    # We are the spawner wait for kids ...
	    push(@kids,$pid);
	    prn "Pressure $P spawned, waiting 10s ...\n";
	    sleep 10;
	} elsif (defined $pid) {
	    # We are the kid, do the job ...
	    runVASPforP $P;
	    exit(0);
	} else {
	    die "Cannot fork: $!\n";
	}
    }
}

prn "\n Workers spawned waiting for finished jobs \n";


while (scalar(@kids)) {
my(@newkids);
							       
foreach my $kid (@kids) {
     if (waitpid($kid,WNOHANG)==$kid) {
          prn "Worker $kid finished\n";
       }
     else {
        push(@newkids,$kid);
     }
}
@kids=@newkids;
#print qq{parent: I have currently }.scalar(@newkids).qq{ kids\n};
sleep(5);
}


prn "Done\n";

#type out the P(V) function
#foreach $v (sort keys %PvsV) {
#    ($p,$e)=split(":",$PvsV{$v});
#    print "$v \t $p \t $e\n";
#}


__END__

$Log: bp,v $
Revision 1.7  2002/06/20 12:38:04  jochym
NSW_DFRM param for optymization in deform added.

Revision 1.6  2002/05/14 07:16:37  jochym
Not set NSW for deformation calc.

Revision 1.5  2002/05/08 13:48:49  jochym
Typo (syntax) corrected

Revision 1.4  2002/05/08 13:45:09  jochym
New ADDGRID parameter and LREAL removed.

Revision 1.3  2002/05/08 13:20:29  jochym
CUSTOM_INCAR added to PARAMS

Revision 1.2  2002/03/22 19:55:17  jochym
Addgrid flag added to INCARS

Revision 1.1  2002/02/22 10:00:20  jochym
Cij added to the repository

Revision 1.18  2001/08/24 15:56:34  jochym
GRAVE error corrected. nsw was set to 30 for single point calcs!!

AIIIIIEEE!

Revision 1.17  2001/08/15 10:33:29  jochym
Reading cell vectors from CONTCAR added (first take - not tested yet).

Revision 1.16  2001/08/14 16:07:37  jochym
Monoclinic deformations added

Revision 1.15  2001/08/08 07:45:51  jochym
Modification to support vasp445. Different spelling of orthorhombic.

Revision 1.14  2001/04/26 17:58:12  jochym
Change GPa to kB in messages

Revision 1.13  2001/04/25 15:04:59  jochym
Pelny przebieg dla vaspa gotowy. Wlacznie z ekstrakcja E,P,V

Revision 1.12  2001/03/27 14:51:23  jochym
Cosmethics

Revision 1.11  2001/03/27 14:28:22  jochym
Lots of clean up. New VASP running logic. Messages.
General deformations procedures (still to be tested).

Revision 1.10  2001/03/17 15:21:48  jochym
First computing (?) version

Revision 1.9  2001/03/17 12:12:11  jochym
Code realizing deformations. First take. (Orthogonal for now).

Revision 1.8  2001/03/15 13:44:52  jochym
Poprawka == vs. eq :(

Revision 1.7  2001/03/14 14:30:45  jochym
First running version. Not working :) yet

Revision 1.6  2001/03/14 13:45:28  jochym
Part of calculations engine started working

Revision 1.5  2001/03/13 22:15:47  jochym
*** empty log message ***

Revision 1.4  2001/03/13 21:01:14  jochym
Kolejny dzien sluzby ... Dodane szkielety obliczen z deformacjami
rozpoznawanie sieci i pare innych drobiazgow. Finalny (?) ksztalt
drzewa katalogow.

Revision 1.3  2001/03/13 13:59:06  jochym
Dalszy ciag VASPizacji

Revision 1.2  2001/03/09 16:33:55  jochym
*** empty log message ***

Revision 1.1  2001/03/09 14:48:55  jochym
Reorganization of the archive

