#!/usr/bin/perl 

use Getopt::Long;

#######################################################
# Procedure definitions
#######################################################

$rev='$Revision: 1.2 $';
$Pi=3.14159265359;

sub usage () {
}

GetOptions(
    'debug|d'		=> \$debug,
    'verbose|v'		=> \$verb,
    'first|f'		=> \$first
) or usage;


sub prtHello{
    print STDERR "Cij (VASP) $rev \n";
}

sub dprn
{
    print STDERR @_ if $debug;
}

sub prn
{
    print STDERR @_ ;
}

sub vprn
{
    print STDERR @_ if $verb;
}

sub getResults {
    my ($fname)=@_;
    my $st=100;
    my ($p, $e, $v);
    my $s={};
    my $g={};
    
    open(CST,$fname) or die "Cannot open $fname:";
    while (<CST>) {
	if ($st==100 & /TOTEN += +([-0-9.]+) eV/) {
	    $st=101;
	    $e=$1;
	}
	if ($st==4 & /^ +([-0-9.]+) +([-0-9.]+) +([-0-9.]+)/) {
	    ${$g}{"a"}=$1;
	    ${$g}{"b"}=$2;
	    ${$g}{"c"}=$3;
	    if ($first){ $st=5; } else { $st=100; }
	}
	if ($st==303 & /^ +length of vectors/){
	    $st=4;
	}
	if ($st==2 & /^ +direct lattice vectors/) {
	    $st=3;
	}
	if ($st==302 & /^ +([-0-9.]+) +([-0-9.]+) +([-0-9.]+)/) {
		$st=303;
		${$g}{"cx"}=$1; ${$g}{"cy"}=$2; ${$g}{"cz"}=$3;
	}
	if ($st==301 & /^ +([-0-9.]+) +([-0-9.]+) +([-0-9.]+)/) {
		$st=302;
		${$g}{"bx"}=$1; ${$g}{"by"}=$2; ${$g}{"bz"}=$3;
	}
	if ($st==3 & /^ +([-0-9.]+) +([-0-9.]+) +([-0-9.]+)/) {
		$st=301;
		${$g}{"ax"}=$1; ${$g}{"ay"}=$2; ${$g}{"az"}=$3;
	}
	if ($st==1 & /^ +volume of cell : +([0-9.]+)/) {
	    $st=2;
	    $v=$1;
	}
	if ($st==101 & /^ +in kB +([-0-9.]+) +([-0-9.]+) +([-0-9.]+) +([-0-9.]+) +([-0-9.]+) +([-0-9.]+)/) {
	    dprn "Stress: $1 $2 $3 $4 $5 $6 \n";
	    ${$s}{"xx"}=$1;
	    ${$s}{"yy"}=$2;
	    ${$s}{"zz"}=$3;
	    ${$s}{"yx"}=${$s}{"xy"}=$4;
	    ${$s}{"yz"}=${$s}{"zy"}=$5;
	    ${$s}{"zx"}=${$s}{"xz"}=$6;
	    $st=102;
	}
	if ($st==102 & /^ +external pressure = +([-0-9.]+) kB  Pullay stress = +([-0-9.]+) kB/) {
	    $p=$1+$2;
	    $st=1;
	}
    }
    close(CST);
#    print STDERR "}",%$g,"{\n";
    ($v, $e, $p, $g, $s);
}

sub printST {
    my (%st)=@_;
    print  STDERR "---------------------------------\n";
    print  STDERR "       x          y          z\n";
    printf STDERR "x %8.2f   %8.2f   %8.2f\n", $st{xx}, $st{xy}, $st{xz};
    printf STDERR "y %8.2f   %8.2f   %8.2f\n", $st{yx}, $st{yy}, $st{yz};
    printf STDERR "z %8.2f   %8.2f   %8.2f\n", $st{zx}, $st{zy}, $st{zz};
    printf STDERR "---------------------------------\n";
}

sub mkStressVect {
    my (%s)=@_;
    [$s{xx},$s{yy},$s{zz},($s{xy}+$s{yx})/2,($s{xz}+$s{zx})/2,($s{yz}+$s{zy})/2];
}

sub prtMtrx {
    my (@m)=@_;
    my ($v, $e);

    foreach $v (@m) {
	print STDERR "[ ";
	foreach $e (@$v) {
	    printf STDERR " %6.3f ", $e;
	}
	print STDERR " ]\n";
    }
}

#
# WARNING !
# I have implemented below code for any deformations
# but it is checked for orthogonal unit cells (so far).
# 
# The formula used for transformation: 
# (A,B,C,alpha,beta,gamma)->(x,y,z)
# will produce incorrect results (sign of C_z coordinate)
# when crossing cos(beta+gamma)=cos(alpha) boundary.
# In normal situations this is not the case, but beware.
#
# This uses following definition of strain:
# For deformation $u_i=x'_i - x_i$ i,j-element of strain
# tensor is: 
# \[
# u_{ij}=\frac{1}{2}\left(\frac{\partial u_i}{\partial x_j}+
#			  \frac{\partial u_j}{\partial x_i}\right)
# \]
# The mkDeformVect function computes only cartesian components
# of the deformation vector. The differentiation and symmetrization
# is left to the Mathematica part.
# The function latMtrx calculates carthesian coordinates of
# cell vectors starting from (A,B,C,alpha,beta,gamma) list.
#

sub csc{
    my ($x)=@_;
    1./sin($x);
}

sub tan{
    my ($x)=@_;
    sin($x)/cos($x);
}


sub cot{
    my ($x)=@_;
    cos($x)/sin($x);
}


sub latMtrx {
    my ($x, $y, $z, $a, $b, $g)=@_;
    my $t;
    
    print STDERR join(", ", @_),"\n" if ($debug>5);
    
    $a*=($Pi/180);
    $b*=($Pi/180);
    $g*=($Pi/180);
    $t=1 - (cot($b)*cot($g)-cos($a)*csc($b)*csc($g))**2;
    if ($t < 0.) {
        $a/=($Pi/180);
	$b/=($Pi/180);
	$g/=($Pi/180);
	print STDERR 
	"__________________________________\n".
	"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n".
	"!!!!!!! YOU ARE IN TROUBLE !!!!!!!\n".
	"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n".
	"__________________________________\n".
	"The parameters (alpha,beta,gamma)=($a,$b,$g)\n".
	"read from the $NowProcessing file are probably incorrect and lead to\n".
	"imaginary coordinates. This range of parameters is unsupported\n".
	"by this program (and is, let me say that, very strange for a crystal).\n".
	"Cennot continue, bye.\n";
	exit 1; 
    }
    (	[$x, 0, 0], 
	[$y*cos($g), $y*sin($g), 0],
	[$z*cos($b), 
	 $z*(cos($a)/sin($g)-cos($b)*cot($g)),
	 $z*sin($b)*sqrt($t)]
    );
#	 
#  This was a brute-force formula for C_y, C_z, the one above
#  should be much better
#
#	    ($z*sqrt((-2*cos(($Pi*$a)/180.) 
#			+ cos(($Pi*($b - $g))/180.)
#			+ cos(($Pi*($b + $g))/180.)
#		     )**2 * csc(($Pi*$g)/180.)**2))/2.,
#	    
#	 $z*sin($b)*
#	 sqrt(1 - (cos($a) - cos($b)*cos($g))**2 *csc($b)**2 * (csc($g))**2)] 
#
}

sub mkDeformVect {
    my ($g,$go)=@_;
    my $a, $b, $c, $ao, $bo, $co;
    my $alf, $bet, $gam, $alfo, $beto, $gamo;
    my $i, $j, @m, @mo, @u;
    my @k=sort keys %{$g};
    my @ko=sort keys %{$go};
    

#   If we have lattice matrix use it
#   Detection by checking if proper indexes exists    

#   First for base lattice
    if (grep(/ax/,@ko) && grep(/by/,@ko) && grep(/cz/,@ko)) {
	@mo=( [ ${$go}{ax}, ${$go}{ay}, ${$go}{az} ],
	      [ ${$go}{bx}, ${$go}{by}, ${$go}{bz} ],
	      [ ${$go}{cx}, ${$go}{cy}, ${$go}{cz} ] );
    } else {
	$ao=${$go}{a};
	$bo=${$go}{b};
	$co=${$go}{c};
	$alfo=${$go}{alpha};
	$beto=${$go}{beta};
	$gamo=${$go}{gamma};
	@mo=latMtrx($ao, $bo, $co, $alfo, $beto, $gamo);    prtMtrx(@mo) if ($debug>2);
    }
    
#   For deformed lattice
    if (grep(/ax/,@k) && grep(/by/,@k) && grep(/cz/,@k)) {
	@m=( [ ${$g}{ax}, ${$g}{ay}, ${$g}{az} ],
	      [ ${$g}{bx}, ${$g}{by}, ${$g}{bz} ],
	      [ ${$g}{cx}, ${$g}{cy}, ${$g}{cz} ] );
    } else {
        $a=${$g}{a}; 	
	$b=${$g}{b}; 	
	$c=${$g}{c}; 	
	$alf=${$g}{alpha}; 	
	$bet=${$g}{beta}; 	
	$gam=${$g}{gamma}; 	
	@m=latMtrx($a, $b, $c, $alf, $bet, $gam);		prtMtrx(@m) if $debug;
    }	

#   Now the deformation
    foreach $i (0..2) {
	foreach $j (0..2) {
	    $u[$i][$j]=$m[$i][$j]-$mo[$i][$j];
	}
    }
    
    prtMtrx(@u) if $debug;
    (\@mo,\@u);
}

sub numcmp{
    my ($x,$y)=($a,$b);
    $x <=> $y;
}

sub prtResults{
    my ($p,$e,$v,$u,$s)=@_;
    printf "{%7.3f, %12.6f, %7.4f,\n",$p,$e,$v;
    printf "\t{{%12.8f,%12.8f,%12.8f}\n",$$u[0][0],$$u[0][1],$$u[0][2];
    printf "\t,{%12.8f,%12.8f,%12.8f}\n",$$u[1][0],$$u[1][1],$$u[1][2];
    printf "\t,{%12.8f,%12.8f,%12.8f}}\n",$$u[2][0],$$u[2][1],$$u[2][2];
    printf "\t\t,{%12.8f,%12.8f,%12.8f,\n\t\t  %12.8f,%12.8f,%12.8f}}\n",
	    $$s[0],$$s[1],$$s[2],$$s[3],$$s[4],$$s[5];
}



############################################################
# Program logic
############################################################

prtHello;

# $verb=7; $debug=7;

if ($ARGV[0] eq "") {
    die "Usage: Cij <Base_Model_Name>\n";
}

$baseName=$ARGV[0];

print STDERR "Analysing $baseName ...\n";

dprn "Base files:";
foreach $f (<${baseName}_*/IDOF/OUTCAR>) {
    if ($f =~ /${baseName}_([-0-9.]+)/){
        dprn "$f ";
	$bFiles{$1}=$f;
    }
}
dprn "\n";

dprn "Deformations:";
foreach $f (<${baseName}_*/*_*_*_*/OUTCAR>) {
    if ($f =~ /${baseName}_([-0-9.]+)_([SU][ABC])_([-0-9.]+)/) {
        dprn "$f ";
	$dFiles{"$1:$2:$3"}=$f;
    }
}
dprn "\n";

# Read in geometry, stress and other data from the base file
foreach $ap (sort numcmp keys %bFiles) {
    dprn "Base file: $bFiles{$ap}:\n";
    ($v,$e,$p,$g,$s)=getResults($bFiles{$ap});
    dprn join(' ',sort keys %{$g}), "\n";
    print STDERR "$bFiles{$ap}: E= $e eV Po= $ap kB  Pt= $p kB\n" if ($verb>1);
    printST %{$s} if ($verb>1);
    $vo{$ap}=$v;
    $po{$ap}=$p;
    $eo{$ap}=$e;
    $so{$ap}=$s;
    $go{$ap}=$g;
    printST %{$so{$ap}} if ($debug >5);
}

# count pressures and deformations 
foreach $f (sort keys %dFiles) {
    print STDERR "$f\n" if ($debug > 5);
    ($apress,$dtype,$dsiz)=split(":",$f);
    $dtypes{$dtype}++;
    $aps{$apress}++;
}

print STDERR "Pressures [kB]: ", join(", ",sort { $a <=> $b } keys %aps), "\n" if $debug;
print STDERR "P[kB]:\t" if ($verb==0);

$N=scalar(keys %aps);
$nrcol=0;
print "{$N\n";

dprn join(' ',keys %aps);

# For each pressure analyse the deformations
foreach $ap (sort numcmp keys %aps) {
    printf STDERR "\n%5.1f (%5.1f kB)...\n", $ap, $po{$ap} if $verb;
    if ($verb==0) {
	printf STDERR "%5.1f(%5.1f) ", $ap, $po{$ap};
	if ((++$nrcol)>4) {
	    print STDERR "\n\t";
	    $nrcol=0;
	};
    }
    $NowProcessing=$bFiles{$ap};
    dprn "Processing $NowProcessing\n";
    ($mo,$uv)=mkDeformVect($go{$ap},$go{$ap});
    $sv=mkStressVect(%{$so{$ap}});
    print ",{\n ";
    prtResults($po{$ap},$eo{$ap},$vo{$ap},$mo,$sv);
    foreach $dt (sort keys %dtypes) {
	foreach $f (sort keys %dFiles) {
	    next if ( $dFiles{$f} !~ /_([-0-9.]+)_${dt}_([-0-9.]+)/);
	    next if ($1 != $ap);
	    $dsiz=$2;
	    $NowProcessing=$dFiles{$f};
	    print STDERR "$dFiles{$f} \n" if ($debug>6);	    
	    printf STDERR " %s %4.1f:", $dt, $2 if ($verb>1);
	    printf STDERR " %s(%4.1f)", $dt, $2 if ($verb==1);
	    ($v,$e,$p,$g,$s)=getResults($dFiles{$f});
	    printf STDERR " E=%7.5e V=%6.3f P=%4.0f ", $e, $v, $p if ($verb>1);
	    $sv=mkStressVect(%$s);
	    printf STDERR "S=[" . " %4.1f" x 6 . "]\n"  ,@{$sv} if ($verb>1);
	    ($mo,$uv)=mkDeformVect($g,$go{$ap});
	    $results{$f}=[$e,$p,$uv,$sv];
	    print ",";
	    prtResults($p,$e,$v,$uv,$sv);
	    print STDERR "$dFiles{$f}:\nE= $e eV Po= $ap kB  Pt= $p kB\n" if ($debug >5);
	    print STDERR "Stress tensor for deformation $dt of $dsiz \n" if ($debug >5);
	    printST %{$s} if ($debug > 5);
	}
    }
    print "}\n";
}

print "}\n";

print STDERR "\n";

__END__


$Log: cij.pl,v $
Revision 1.2  2002/05/17 16:36:16  jochym
Added support for high and low freq. cij (first and last iter. in optimization)

Revision 1.1  2002/02/22 10:00:20  jochym
Cij added to the repository

Revision 1.6  2001/09/10 15:13:10  jochym
Small corrections in the cij.pl and Cij

Revision 1.5  2001/09/09 20:36:32  jochym
Debug (-d) and verbose (-v) switches added

Revision 1.4  2001/04/27 14:49:50  jochym
Error in mkDeformVect corrected (results now verified for cubic case).
AgBr tests run correctly as well as cubic Si test

Revision 1.3  2001/04/26 13:38:49  jochym
Finished porting to VASP. Not yet tested except on cubic Si (non-physical).
Cij to be corrected for final version - uses cij.pl in current directory.
All to be tested on AgBr (CsCl or NaCl structure)

Revision 1.2  2001/04/25 17:25:23  jochym
First attempts to vaspization of cij.pl. Data collection.

Revision 1.1  2001/03/09 14:48:55  jochym
Reorganization of the archive

Revision 1.1.1.1  2001/03/09 10:11:10  jochym
Cij script. Working CASTEP version. Checkin before branching vasp version.

Revision 1.7  1999/09/22 21:05:21  jochym
Poprawki w wyrazeniach na u (znaczace !). Czyszczenie.

Revision 1.6  1999/09/21 04:45:02  jochym
prierwsza liczaca wersja. Ciagle klopoty z wynikami.
Niesprawdzone znaki.

Revision 1.5  1999/09/20 22:14:44  jochym
Po edycji w pracy

Revision 1.4  1999/09/17 16:29:26  jochym
Poprawiony blad we wzorze na kartezjanskie wsp. wektorow sieci.

Revision 1.3  1999/09/13 08:48:43  jochym
po edycji w domu

Revision 1.4  1999/09/13 05:11:30  jochym
Po edycji w domu

Revision 1.3  1999/09/10 07:36:12  jochym
instytut, przed zmianami w domu.

Revision 1.2  1999/09/07 14:14:16  jochym
kolejny etap analizatora. Koniec dnia.

Revision 1.1  1999/09/07 07:08:53  jochym
Initial revision

Revision 1.1  1999/09/06 19:32:05  jochym
Initial revision


